from collections import defaultdict
from functools import partial
from itertools import chain
from typing import (
    Any, Callable, Dict, List, Optional, Union, Tuple, cast)

from ..error import GraphQLError
from ..language import (
    DirectiveDefinitionNode, DocumentNode,
    EnumTypeDefinitionNode, EnumTypeExtensionNode,
    InputObjectTypeDefinitionNode, InputObjectTypeExtensionNode,
    InterfaceTypeDefinitionNode, InterfaceTypeExtensionNode,
    ObjectTypeDefinitionNode, ObjectTypeExtensionNode, OperationType,
    ScalarTypeDefinitionNode, ScalarTypeExtensionNode,
    SchemaExtensionNode, SchemaDefinitionNode,
    UnionTypeDefinitionNode, UnionTypeExtensionNode,
    NamedTypeNode, TypeExtensionNode)
from ..type import (
    GraphQLArgument, GraphQLArgumentMap, GraphQLDirective,
    GraphQLEnumType, GraphQLEnumValue, GraphQLEnumValueMap,
    GraphQLField, GraphQLFieldMap, GraphQLInputField, GraphQLInputFieldMap,
    GraphQLInputObjectType, GraphQLInputType, GraphQLInterfaceType,
    GraphQLList, GraphQLNamedType, GraphQLNonNull, GraphQLObjectType,
    GraphQLScalarType, GraphQLSchema, GraphQLType, GraphQLUnionType,
    is_enum_type, is_input_object_type, is_interface_type, is_list_type,
    is_non_null_type, is_object_type, is_scalar_type, is_schema, is_union_type,
    is_introspection_type, is_specified_scalar_type)
from .build_ast_schema import ASTDefinitionBuilder

__all__ = ['extend_schema']


def extend_schema(schema: GraphQLSchema, document_ast: DocumentNode,
                  assume_valid=False) -> GraphQLSchema:
    """Extend the schema with extensions from a given document.

    Produces a new schema given an existing schema and a document which may
    contain GraphQL type extensions and definitions. The original schema will
    remain unaltered.

    Because a schema represents a graph of references, a schema cannot be
    extended without effectively making an entire copy. We do not know until
    it's too late if subgraphs remain unchanged.

    This algorithm copies the provided schema, applying extensions while
    producing the copy. The original schema remains unaltered.

    When extending a schema with a known valid extension, it might be safe to
    assume the schema is valid. Set `assume_valid` to true to assume the
    produced schema is valid.
    """

    if not is_schema(schema):
        raise TypeError('Must provide valid GraphQLSchema')

    if not isinstance(document_ast, DocumentNode):
        'Must provide valid Document AST'

    # Collect the type definitions and extensions found in the document.
    type_definition_map: Dict[str, Any] = {}
    type_extensions_map: Dict[str, Any] = defaultdict(list)

    # New directives and types are separate because a directives and types can
    # have the same name. For example, a type named "skip".
    directive_definitions: List[DirectiveDefinitionNode] = []

    # Schema extensions are collected which may add additional operation types.
    schema_extensions: List[SchemaExtensionNode] = []

    for def_ in document_ast.definitions:
        if isinstance(def_, SchemaDefinitionNode):
            # Sanity check that a schema extension is not defining a new schema
            raise GraphQLError(
                'Cannot define a new schema within a schema extension.',
                [def_])
        elif isinstance(def_, SchemaExtensionNode):
            schema_extensions.append(def_)
        elif isinstance(def_, (
                ObjectTypeDefinitionNode,
                InterfaceTypeDefinitionNode,
                EnumTypeDefinitionNode,
                UnionTypeDefinitionNode,
                ScalarTypeDefinitionNode,
                InputObjectTypeDefinitionNode)):
            # Sanity check that none of the defined types conflict with the
            # schema's existing types.
            type_name = def_.name.value
            if schema.get_type(type_name):
                raise GraphQLError(
                    f"Type '{type_name}' already exists in the schema."
                    ' It cannot also be defined in this type definition.',
                    [def_])
            type_definition_map[type_name] = def_
        elif isinstance(def_, (
                ScalarTypeExtensionNode,
                ObjectTypeExtensionNode,
                InterfaceTypeExtensionNode,
                EnumTypeExtensionNode,
                InputObjectTypeExtensionNode,
                UnionTypeExtensionNode)):
            # Sanity check that this type extension exists within the
            # schema's existing types.
            extended_type_name = def_.name.value
            existing_type = schema.get_type(extended_type_name)
            if not existing_type:
                raise GraphQLError(
                    f"Cannot extend type '{extended_type_name}'"
                    ' because it does not exist in the existing schema.',
                    [def_])
            check_extension_node(existing_type, def_)
            type_extensions_map[extended_type_name].append(def_)
        elif isinstance(def_, DirectiveDefinitionNode):
            directive_name = def_.name.value
            existing_directive = schema.get_directive(directive_name)
            if existing_directive:
                raise GraphQLError(
                    f"Directive '{directive_name}' already exists"
                    ' in the schema. It cannot be redefined.', [def_])
            directive_definitions.append(def_)

    # If this document contains no new types, extensions, or directives then
    # return the same unmodified GraphQLSchema instance.
    if (not type_extensions_map and not type_definition_map
            and not directive_definitions and not schema_extensions):
        return schema

    # Below are functions used for producing this schema that have closed over
    # this scope and have access to the schema, cache, and newly defined types.

    def get_merged_directives() -> List[GraphQLDirective]:
        if not schema.directives:
            raise TypeError('schema must have default directives')

        return list(chain(
            map(extend_directive, schema.directives),
            map(ast_builder.build_directive, directive_definitions)))

    def extend_maybe_named_type(
            type_: Optional[GraphQLNamedType]) -> Optional[GraphQLNamedType]:
        return extend_named_type(type_) if type_ else None

    def extend_named_type(type_: GraphQLNamedType) -> GraphQLNamedType:
        if is_introspection_type(type_) or is_specified_scalar_type(type_):
            # Builtin types are not extended.
            return type_

        name = type_.name
        if name not in extend_type_cache:
            if is_scalar_type(type_):
                type_ = cast(GraphQLScalarType, type_)
                extend_type_cache[name] = extend_scalar_type(type_)
            elif is_object_type(type_):
                type_ = cast(GraphQLObjectType, type_)
                extend_type_cache[name] = extend_object_type(type_)
            elif is_interface_type(type_):
                type_ = cast(GraphQLInterfaceType, type_)
                extend_type_cache[name] = extend_interface_type(type_)
            elif is_enum_type(type_):
                type_ = cast(GraphQLEnumType, type_)
                extend_type_cache[name] = extend_enum_type(type_)
            elif is_input_object_type(type_):
                type_ = cast(GraphQLInputObjectType, type_)
                extend_type_cache[name] = extend_input_object_type(type_)
            elif is_union_type(type_):
                type_ = cast(GraphQLUnionType, type_)
                extend_type_cache[name] = extend_union_type(type_)

        return extend_type_cache[name]

    def extend_directive(directive: GraphQLDirective) -> GraphQLDirective:
        return GraphQLDirective(
            directive.name,
            description=directive.description,
            locations=directive.locations,
            args=extend_args(directive.args),
            ast_node=directive.ast_node)

    def extend_input_object_type(
            type_: GraphQLInputObjectType) -> GraphQLInputObjectType:
        name = type_.name
        extension_ast_nodes = (
                list(type_.extension_ast_nodes) + type_extensions_map[name]
                if type_.extension_ast_nodes else type_extensions_map[name]
            ) if name in type_extensions_map else type_.extension_ast_nodes
        return GraphQLInputObjectType(
            name,
            description=type_.description,
            fields=lambda: extend_input_field_map(type_),
            ast_node=type_.ast_node,
            extension_ast_nodes=extension_ast_nodes)

    def extend_input_field_map(
            type_: GraphQLInputObjectType) -> GraphQLInputFieldMap:
        old_field_map = type_.fields
        new_field_map = {field_name: GraphQLInputField(
            cast(GraphQLInputType, extend_type(field.type)),
            description=field.description,
            default_value=field.default_value,
            ast_node=field.ast_node)
            for field_name, field in old_field_map.items()}

        # If there are any extensions to the fields, apply those here.
        extensions = type_extensions_map.get(type_.name)
        if extensions:
            for extension in extensions:
                for field in extension.fields:
                    field_name = field.name.value
                    if field_name in old_field_map:
                        raise GraphQLError(
                            f"Field '{type_.name}.{field_name}' already"
                            ' exists in the schema. It cannot also be defined'
                            ' in this type extension.', [field])
                    new_field_map[field_name] = ast_builder.build_input_field(
                        field)

        return new_field_map

    def extend_enum_type(type_: GraphQLEnumType) -> GraphQLEnumType:
        name = type_.name
        extension_ast_nodes = (
                list(type_.extension_ast_nodes) + type_extensions_map[name]
                if type_.extension_ast_nodes else type_extensions_map[name]
            ) if name in type_extensions_map else type_.extension_ast_nodes
        return GraphQLEnumType(
            name,
            description=type_.description,
            values=extend_value_map(type_),
            ast_node=type_.ast_node,
            extension_ast_nodes=extension_ast_nodes)

    def extend_value_map(type_: GraphQLEnumType) -> GraphQLEnumValueMap:
        old_value_map = type_.values
        new_value_map = {value_name: GraphQLEnumValue(
            value.value,
            description=value.description,
            deprecation_reason=value.deprecation_reason,
            ast_node=value.ast_node)
            for value_name, value in old_value_map.items()}

        # If there are any extensions to the values, apply those here.
        extensions = type_extensions_map.get(type_.name)
        if extensions:
            for extension in extensions:
                for value in extension.values:
                    value_name = value.name.value
                    if value_name in old_value_map:
                        raise GraphQLError(
                            f"Enum value '{type_.name}.{value_name}' already"
                            ' exists in the schema. It cannot also be defined'
                            ' in this type extension.', [value])
                    new_value_map[value_name] = ast_builder.build_enum_value(
                        value)

        return new_value_map

    def extend_scalar_type(type_: GraphQLScalarType) -> GraphQLScalarType:
        name = type_.name
        extension_ast_nodes = (
                list(type_.extension_ast_nodes) + type_extensions_map[name]
                if type_.extension_ast_nodes else type_extensions_map[name]
            ) if name in type_extensions_map else type_.extension_ast_nodes
        return GraphQLScalarType(
            name,
            serialize=type_.serialize,
            description=type_.description,
            parse_value=type_.parse_value,
            parse_literal=type_.parse_literal,
            ast_node=type_.ast_node,
            extension_ast_nodes=extension_ast_nodes)

    def extend_object_type(type_: GraphQLObjectType) -> GraphQLObjectType:
        name = type_.name
        extension_ast_nodes = type_.extension_ast_nodes
        try:
            extensions = type_extensions_map[name]
        except KeyError:
            pass
        else:
            if extension_ast_nodes:
                extension_ast_nodes = list(
                    extension_ast_nodes) + extensions
            else:
                extension_ast_nodes = extensions
        return GraphQLObjectType(
            type_.name,
            description=type_.description,
            interfaces=partial(extend_implemented_interfaces, type_),
            fields=partial(extend_field_map, type_),
            ast_node=type_.ast_node,
            extension_ast_nodes=extension_ast_nodes,
            is_type_of=type_.is_type_of)

    def extend_args(args: GraphQLArgumentMap) -> GraphQLArgumentMap:
        return {arg_name: GraphQLArgument(
            cast(GraphQLInputType, extend_type(arg.type)),
            default_value=arg.default_value,
            description=arg.description,
            ast_node=arg.ast_node)
            for arg_name, arg in args.items()}

    def extend_interface_type(
            type_: GraphQLInterfaceType) -> GraphQLInterfaceType:
        name = type_.name
        extension_ast_nodes = type_.extension_ast_nodes
        try:
            extensions = type_extensions_map[name]
        except KeyError:
            pass
        else:
            if extension_ast_nodes:
                extension_ast_nodes = list(
                    extension_ast_nodes) + extensions
            else:
                extension_ast_nodes = extensions
        return GraphQLInterfaceType(
            type_.name,
            description=type_.description,
            fields=partial(extend_field_map, type_),
            ast_node=type_.ast_node,
            extension_ast_nodes=extension_ast_nodes,
            resolve_type=type_.resolve_type)

    def extend_union_type(type_: GraphQLUnionType) -> GraphQLUnionType:
        name = type_.name
        extension_ast_nodes = (
                list(type_.extension_ast_nodes) + type_extensions_map[name]
                if type_.extension_ast_nodes else type_extensions_map[name]
            ) if name in type_extensions_map else type_.extension_ast_nodes
        return GraphQLUnionType(
            name,
            description=type_.description,
            types=lambda: extend_possible_types(type_),
            ast_node=type_.ast_node,
            resolve_type=type_.resolve_type,
            extension_ast_nodes=extension_ast_nodes)

    def extend_possible_types(
            type_: GraphQLUnionType) -> List[GraphQLObjectType]:
        possible_types = list(map(extend_named_type, type_.types))

        # If there are any extensions to the union, apply those here.
        extensions = type_extensions_map.get(type_.name)
        if extensions:
            for extension in extensions:
                for named_type in extension.types:
                    # Note: While this could make early assertions to get the
                    # correctly typed values, that would throw immediately
                    # while type system validation with validate_schema() will
                    # produce more actionable results.
                    possible_types.append(ast_builder.build_type(named_type))

        return cast(List[GraphQLObjectType], possible_types)

    def extend_implemented_interfaces(
            type_: GraphQLObjectType) -> List[GraphQLInterfaceType]:
        interfaces: List[GraphQLInterfaceType] = list(
            map(cast(Callable[[GraphQLNamedType], GraphQLInterfaceType],
                     extend_named_type), type_.interfaces))

        # If there are any extensions to the interfaces, apply those here.
        for extension in type_extensions_map[type_.name]:
            for named_type in extension.interfaces:
                # Note: While this could make early assertions to get the
                # correctly typed values, that would throw immediately while
                # type system validation with validate_schema() will produce
                # more actionable results.
                interfaces.append(
                    cast(GraphQLInterfaceType, build_type(named_type)))

        return interfaces

    def extend_field_map(
                type_: Union[GraphQLObjectType, GraphQLInterfaceType]
            ) -> GraphQLFieldMap:
        old_field_map = type_.fields
        new_field_map = {field_name: GraphQLField(
            cast(GraphQLObjectType, extend_type(field.type)),
            description=field.description,
            deprecation_reason=field.deprecation_reason,
            args=extend_args(field.args),
            ast_node=field.ast_node,
            resolve=field.resolve)
            for field_name, field in old_field_map.items()}

        # If there are any extensions to the fields, apply those here.
        for extension in type_extensions_map[type_.name]:
            for field in extension.fields:
                field_name = field.name.value
                if field_name in old_field_map:
                    raise GraphQLError(
                        f"Field '{type_.name}.{field_name}'"
                        ' already exists in the schema.'
                        ' It cannot also be defined in this type extension.',
                        [field])
                new_field_map[field_name] = build_field(field)

        return new_field_map

    # noinspection PyTypeChecker,PyUnresolvedReferences
    def extend_type(type_def: GraphQLType) -> GraphQLType:
        if is_list_type(type_def):
            return GraphQLList(extend_type(type_def.of_type))  # type: ignore
        if is_non_null_type(type_def):
            return GraphQLNonNull(  # type: ignore
                extend_type(type_def.of_type))  # type: ignore
        return extend_named_type(type_def)  # type: ignore

    # noinspection PyShadowingNames
    def resolve_type(type_ref: NamedTypeNode) -> GraphQLNamedType:
        type_name = type_ref.name.value
        existing_type = schema.get_type(type_name)
        if existing_type:
            return extend_named_type(existing_type)
        raise GraphQLError(
            f"Unknown type: '{type_name}'."
            ' Ensure that this type exists either in the original schema,'
            ' or is added in a type definition.', [type_ref])

    ast_builder = ASTDefinitionBuilder(
        type_definition_map,
        assume_valid=assume_valid, resolve_type=resolve_type)
    build_field = ast_builder.build_field
    build_type = ast_builder.build_type

    extend_type_cache: Dict[str, GraphQLNamedType] = {}

    # Get the extended root operation types.
    operation_types = {
        OperationType.QUERY: extend_maybe_named_type(schema.query_type),
        OperationType.MUTATION: extend_maybe_named_type(schema.mutation_type),
        OperationType.SUBSCRIPTION:
            extend_maybe_named_type(schema.subscription_type)}

    # Then, incorporate all schema extensions.
    for schema_extension in schema_extensions:
        if schema_extension.operation_types:
            for operation_type in schema_extension.operation_types:
                operation = operation_type.operation
                if operation_types[operation]:
                    raise TypeError(f'Must provide only one {operation.value}'
                                    ' type in schema.')
                type_ref = operation_type.type
                # Note: While this could make early assertions to get the
                # correctly typed values, that would throw immediately while
                # type system validation with validate_schema() will produce
                # more actionable results
                operation_types[operation] = ast_builder.build_type(type_ref)

    schema_extension_ast_nodes = (
        schema.extension_ast_nodes or cast(Tuple[SchemaExtensionNode], ())
        ) + tuple(schema_extensions)

    # Iterate through all types, getting the type definition for each, ensuring
    # that any type not directly referenced by a value will get created.
    types = list(map(extend_named_type, schema.type_map.values()))
    # do the same with new types
    types.extend(ast_builder.build_type(type_)
                 for type_ in type_definition_map.values())

    # Then produce and return a Schema with these types.
    return GraphQLSchema(  # type: ignore
        query=operation_types[OperationType.QUERY],
        mutation=operation_types[OperationType.MUTATION],
        subscription=operation_types[OperationType.SUBSCRIPTION],
        types=types,
        directives=get_merged_directives(),
        ast_node=schema.ast_node,
        extension_ast_nodes=schema_extension_ast_nodes)


def check_extension_node(type_: GraphQLNamedType, node: TypeExtensionNode):
    if isinstance(node, ObjectTypeExtensionNode):
        if not is_object_type(type_):
            raise GraphQLError(
                f"Cannot extend non-object type '{type_.name}'.", [node])
    elif isinstance(node, InterfaceTypeExtensionNode):
        if not is_interface_type(type_):
            raise GraphQLError(
                f"Cannot extend non-interface type '{type_.name}'.", [node])
    elif isinstance(node, EnumTypeExtensionNode):
        if not is_enum_type(type_):
            raise GraphQLError(
                f"Cannot extend non-enum type '{type_.name}'.", [node])
    elif isinstance(node, UnionTypeExtensionNode):
        if not is_union_type(type_):
            raise GraphQLError(
                f"Cannot extend non-union type '{type_.name}'.", [node])
    elif isinstance(node, InputObjectTypeExtensionNode):
        if not is_input_object_type(type_):
            raise GraphQLError(
                f"Cannot extend non-input object type '{type_.name}'.", [node])
